# ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: #
# Date: 07.10.2016                                                            #
# Author: Ole-Johan Skrede                                                    #
#                                                                             #
# Solution proposal as part of the exercise program in                        #
# INF4300 - Digital image analysis at the University of Oslo                  #
#                                                                             #
# ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: #
"""
Solution proposal for exercise 7, task 4 and 5.
"""

# pylint: disable=expression-not-assigned
# pylint: disable=bad-indentation

import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns # pylint: disable=import-error,unused-import

def discrimination_function(feature_vector, class_mean, cov_matrix):
  """
  Computes the discrimination function value given the arguments, for a
  Gaussian classifier with assumed equal prior for each class.
  """

  mean_diff = feature_vector - class_mean
  cov_matrix_inv = np.linalg.inv(cov_matrix)
  gauss_kernel = -1/2*np.dot(mean_diff.T, np.dot(cov_matrix_inv, mean_diff))

  return -1/2*np.log(np.linalg.det(cov_matrix)) + gauss_kernel

def main():
  """main"""

  print('='*80)
  print('Solution to weekly exercises in INF4300')
  print('Exercise 7')
  print('Task 4 and 5')
  print('-'*80)

  which_task = 'task 4'

  # Task 4 a) or 5
  if which_task == 'task 4':
    sigma = np.array([[1.2, 0.4], [0.4, 1.8]])
    det_sigma = 2
    sigma_inv = np.array([[1.8, -0.4], [-0.4, 1.2]]) / det_sigma
    feature_val = np.array([1.6, 1.5])
    class_means = [np.array([0.1, 0.1]), np.array([2.1, 1.9]), np.array([-1.5, 2.0])]
  elif which_task == 'task 5':
    sigma = np.array([[1.1, 0.3], [0.3, 1.9]])
    sigma_inv = np.array([[0.95, -0.15], [-0.15, 1.55]])
    feature_val = np.array([1.0, 2.2])
    class_means = [np.array([0, 0]), np.array([3, 3])]


  disc_vals = []
  for class_mean in class_means:
    disc_vals.append(discrimination_function(feature_val, class_mean, sigma))

  print('Vector {0} belongs in class {1:.0f}'.format(feature_val, 1+np.argmax(disc_vals)))


  # Task 4 b)
  if which_task == 'task 4':
    class_mean = np.array([2.1, 1.9])

    # Eigenvalues and eigenvectors of the covariance matrix
    #eig_vals, eig_vecs = np.linalg.eig(sigma)

    delta = 0.025
    xmin = -2
    xmax = 8
    ymin = -2
    ymax = 8
    x_vals = np.arange(xmin, xmax, delta)
    y_vals = np.arange(ymin, ymax, delta)
    xx, _ = np.meshgrid(x_vals, y_vals)
    mahalanobis = np.zeros_like(xx)
    for i, x in enumerate(x_vals):
      for j, y in enumerate(y_vals):
        mean_diff = np.array([x, y]) - class_mean
        # Note j, i because of pyplot imshow convention
        mahalanobis[j, i] = np.sqrt(np.dot(mean_diff.T, np.dot(sigma_inv, mean_diff)))

    plt.figure()
    im = plt.imshow(mahalanobis, interpolation='bilinear', origin='lower',
                    cmap='magma', extent=[xmin, xmax, ymin, ymax])
    levels = np.arange(0, 8, 1)
    CS = plt.contour(mahalanobis, levels, origin='lower', linewidths=2,
                     colors='white', extent=[xmin, xmax, ymin, ymax])

    plt.clabel(CS, levels[1::2], inline=1, fmt='%1.1f', fontsize=14)
    plt.colorbar(im)
    plt.xlabel('x')
    plt.ylabel('y')
    plt.grid('off')
    plt.tight_layout()

    plt.show()


if __name__ == "__main__":
  main()
